#ifndef __FORM_H
#define __FORM_H

#ifndef __PRINTER_H
#include <printer.h>
#endif

#ifndef __SCANNER_H
#include <scanner.h>
#endif

#ifndef __MASK_H
class TMask;
#endif

#ifndef __ISAM_H
#include <isam.h>
#endif

#ifndef __RELATION_H
class TRelation;
class TRelation_description;
class TCursor;
#endif

enum pagetype { odd_page, even_page, first_page, last_page };

class TForm;
class TForm_item;

class TPrint_section : public TArray
{
  static TMask* _msk;

  word _height;           // Altezza della sezione
  int _x, _y;             // Offset di stampa   

  TForm* _form;           // Form cui appartiene alla sezione
  pagetype _page_type;    // Tipo della pagina
  
  TArray _item;           // Lista dei campi da stampare
  
  const TPrint_section& copy(const TPrint_section& ps);
  
protected:
  virtual void print_on(ostream& out) const;
  virtual TForm_item* parse_item(const TString& key);
  TForm_item* parse_item(TScanner& scanner);

public:
  TPrintrow& row(int num);
  TForm& form() const { return *_form; }
  
  TForm_item& field(int n) const { return (TForm_item&)_item[n]; }
  TForm_item& find_field(short id) const;

  word fields() const { return _item.items(); }
  word height() const { return _height; }
  int offset_x() const { return _x; }
  int offset_y() const { return _y; }
  void offset(int& x, int& y) const;
  
  virtual bool ok() const { return height() > 0 || fields() > 0; }
  
  void reset();
  virtual bool update();
  bool parse(TScanner& scanner);

  bool edit(const char* title, bool all);
  pagetype page_type() const { return _page_type; }
  
  const TPrint_section& operator=(const TPrint_section& ps) { return copy(ps); }
  TPrint_section(TForm* parent, pagetype pt);
  TPrint_section(const TPrint_section& ps) { copy(ps); }
  virtual ~TPrint_section();
};


class TForm : public TObject
{                    
  friend class TForm_editor;

  TFilename _name;        // Profile name                     
  TString16 _code;        // Profile code                    
  
  TLocalisamfile _forms;  // forms names file
  TLocalisamfile _prof;   // profile definition file

  TRelation* _relation;   // Can be NULL
  TCursor* _cursor;       // Can be NULL
  TRelation_description* _reldef; // pure

  TArray _head;           // Headers
  TArray _body;           // Bodies
  TArray _foot;           // Footers
  TArray _back;           // Graphic background

  char    _permissions[9]; // user permissions
  bool    _lastpage;       // I am about to print the last page?

  bool    _isnew;          // new form
  bool    _isbase;         // base form (.frm file) 
  TString _desc;           // form description
  
  TPrint_section* exist(char s, pagetype t, bool create = FALSE);   // Can be NULL
  
  static void header_handler(TPrinter& p);
  static void footer_handler(TPrinter& p);                                         

protected:                                                
  // H = Header, B = Body, F = Footer, R = Relation
  TPrint_section& section(char s = 'B', word page = 1);
  word height();          // Height of first page
  
  void print_section(ostream& out, char s) const;
  virtual void print_on(ostream& out) const;
  bool edit_relation();

  bool parse_use(TScanner&);
  bool parse_join(TScanner&);
  bool parse_permissions(TScanner&);
  bool parse_descfield(TScanner&);
  
  bool read_profile();
  bool write_profile();
  
  word page(const TPrinter& p) const;
  
  virtual long records() const;
  virtual word set_background(word p, bool u);
  virtual word set_header(word p, bool u);
  virtual word set_body(word p, bool u);
  virtual word set_footer(word p, bool u);
  
public:               

  bool has_add_permission()      const { return _permissions[0] == 'A'; }
  bool has_remove_permission()   const { return _permissions[1] == 'R'; }
  bool has_modify_permission()   const { return _permissions[2] == 'M'; }
  bool has_relation_permission() const { return _permissions[3] == 'L'; }

  bool print(long from = 0L, long to = -1L);
  
  const TFilename& name() const { return _name; }
  
  TRelation* relation() const { return _relation; }
  TRelation_description* reldef() const { return _reldef; }

  TCursor* cursor() const { return _cursor; }
  virtual const char* validate(const char* cur, TToken_string& val);
  
  TForm_item& find_field(char sec, pagetype pag, short id) const;
  
  // if code == NULL it's a base form
  // otherwise it's integrated by a file definition                                       
  TForm(const char* form, const char* code = NULL);
  virtual ~TForm();
};

#endif
